import pandas as pd
from sklearn.feature_selection import RFE, RFECV


def rfe(dataframe, label_column, estimator, n_features_to_select,**kwargs):

    """ This function is used to perform Recursive Feature Elimination.
    Args:
        estimator (str): The estimator to use for feature selection. Can be one of SVC(),SVR(), RandomForestClassifier(), 
        RandomForestRegressor(), GradientBoostingClassifier(), GradientBoostingRegressor(), LogisticRegression(), LinearRegression(), etc..
        n_features_to_select (int): The number of features to select.
        step (int): The number of features to remove at each iteration.
        verbose (int): Controls verbosity of output.
    Returns:
        pandas.DataFrame: The DataFrame containing the selected features.
    """
    
    # Perform Recursive Feature Elimination
    selector = RFE(estimator=estimator, n_features_to_select=n_features_to_select, **kwargs)
    feature_array = selector.fit_transform(dataframe.drop([label_column], axis =1), dataframe[label_column])
    best_columns= list(selector.get_feature_names_out())
    best_columns.append('Survived')
    df = dataframe[best_columns]

    return df, selector

def rfecv(dataframe, label_column, estimator,scoring, n_jobs = -1, cv=5, **kwargs):

    """ This function is used to perform Recursive Feature Elimination with Cross-Validation.
    Args:
        estimator (str): The estimator to use for feature selection. Import the estimator from sklearn and Can be one of SVC(),SVR(), RandomForestClassifier(), 
        RandomForestRegressor(), GradientBoostingClassifier(), GradientBoostingRegressor(), LogisticRegression(), LinearRegression(), etc..
        step (int): The number of features to remove at each iteration.
    Returns:
        pandas.DataFrame: The DataFrame containing the selected features.
    """
    
    # Perform Recursive Feature Elimination with Cross-Validation
    selector = RFECV(estimator= estimator, scoring= scoring, cv= cv, n_jobs= n_jobs, **kwargs)
    feature_array = selector.fit_transform(dataframe.drop([label_column], axis =1), dataframe[label_column])
    best_columns= list(selector.get_feature_names_out())
    best_columns.append('Survived')
    df = dataframe[best_columns]
    return df, selector



import pandas as pd
import sqlite3
import pymongo
import psycopg2
from mysql import connector



def read_sqlite(query, path_to_db, **kwargs):
    """Reads data from a SQL query and returns a pandas DataFrame.

    Args:
        query (str): The SQL query to execute.
        path_to_db (sqlalchemy.engine.Connection): Path to database "database.db".
        **kwargs: Keyword arguments to pass to pandas.read_sql.

    Returns:
        pandas.DataFrame: The DataFrame containing the data from the SQL query.
    """
    try:
        conn = sqlite3.connect(path_to_db, **kwargs)
        df = pd.read_sql(query, conn, **kwargs)
        conn.close()

    except ConnectionError as conn:
        print("Could not connect to SQLite database.")
        print(conn)

    except Exception as e:
        print("An error occurred.")
        print(e)

    return df

def read_postgres(query, host, port, database_name, username=None, password=None, **kwargs):
    """Reads data from a PostgreSQL database and returns a pandas DataFrame.

    Args:
        query (str): The SQL query to execute.
        host (str): The host of the PostgreSQL database.
        port (int): The port of the PostgreSQL database.
        database_name (str): The name of the PostgreSQL database.
        username (str): The username of the PostgreSQL database. 
        password (str): The password of the PostgreSQL database. 
        **kwargs: Keyword arguments to pass to pandas.read_sql.

    Returns:
        pandas.DataFrame: The DataFrame containing the data from the PostgreSQL database.
    """
    try:
        conn = psycopg2.connect(host=host, port=port, database=database_name, user=username, password=password, **kwargs)
        df = pd.read_sql(query, conn)
        conn.close()

    except ConnectionError as conn:
        print("Could not connect to PostgreSQL database.")
        print(conn)

    except Exception as e:
        print("An error occurred.")
        print(e)

    return df

def read_mysql(query, host, port, database_name, username=None, password=None, **kwargs):
    """Reads data from a MySQL database and returns a pandas DataFrame.

    Args:
        query (str): The SQL query to execute.
        host (str): The host of the MySQL database.
        port (int): The port of the MySQL database.
        database_name (str): The name of the MySQL database.
        username (str): The username of the MySQL database. 
        password (str): The password of the MySQL database. 
        **kwargs: Keyword arguments to pass to pandas.read_sql.

    Returns:
        pandas.DataFrame: The DataFrame containing the data from the MySQL database.
    """
    try:
        conn = connector.connect(host=host, port=port, database=database_name, user=username, password=password, **kwargs)
        df = pd.read_sql(query, conn)
        conn.close()

    except ConnectionError as conn:
        print("Could not connect to MySQL database.")
        print(conn)

    except Exception as e:
        print("An error occurred.")
        print(e)

    return df

def read_mongo(host, port, database_name, collection_name, username=None, password=None, **kwargs):
    """Reads data from a MongoDB database and returns a pandas DataFrame.

    Args:
        host (str): The host of the MongoDB database.
        port (int): The port of the MongoDB database.
        database_name (str): The name of the MongoDB database.
        collection_name (str): The name of the MongoDB collection.
        username (str): The username of the MongoDB database.
        password (str): The password of the MongoDB database.
        **kwargs: Keyword arguments to pass to pandas.read_mongo.

    Returns:
        pandas.DataFrame: The DataFrame containing the data from the MongoDB database.
    """
    try:
        if username and password:
            client = pymongo.MongoClient(host, port, username=username, password=password, **kwargs)
        else:
            client = pymongo.MongoClient(host, port)

        db = client[database_name]
        collection = db[collection_name]

        records = collection.find()

    except ConnectionError as conn:
        print("Could not connect to MongoDB database.")
        print(conn)

    except Exception as e:
        print("An error occurred.")
        print(e)

    return pd.DataFrame(list(records), **kwargs)






from autopilotml.data_loading import read_csv, read_excel, read_sqlite, read_postgres, read_mysql, read_mongo
from autopilotml.preprocessing import drop_missing_rows, drop_missing_columns, imputation, inter_quartile_range, remove_outliers_zscore
from autopilotml.transformation import ordinal_transform, onehot_transform, label_transform
from autopilotml.scaling import standard_scale, target_scale, robust_scale, maxabs_scale, power_transform, quantile_transform
from autopilotml.feature_selection import rfe, rfecv
from autopilotml.model_building import model_fitting
from autopilotml.model_tuning import tuner


# import read_csv, read_excel, read_sqlite, read_postgres, read_mysql, read_mongo
# import drop_missing_rows, drop_missing_columns, imputation, inter_quartile_range, remove_outliers_zscore
# import ordinal_transform, onehot_transform, label_transform
# import standard_scale, target_scale, robust_scale, maxabs_scale, power_transform, quantile_transform
# import rfe, rfecv
# import model_fitting
# import tuner


from sklearn.metrics import mean_squared_error, accuracy_score
import joblib
import optuna
import mlflow
from mlflow import MlflowClient
from datetime import datetime
from sklearn.model_selection import train_test_split
import pandas as pd
import warnings
warnings.filterwarnings("ignore")



#Stage 1: Data Loading

def load_data(path, csv = True, excel = False, **kwargs):
    """ This function is used to load the data from the source.

    Args:
        path (str): The path to the source file.
        csv (bool): If True, the source file is a CSV file.
        excel (bool): If True, the source file is an Excel file.
        **kwargs: Keyword arguments to pass to pandas.read_csv or pandas.read_excel.

    Returns:
        pandas.DataFrame: The DataFrame containing the data.
    """
    # Load the data from the csv source
    if csv:
        df = read_csv(path, **kwargs)
    elif excel:
        df = read_excel(path, **kwargs)

    return df

def load_database(database_type, query, sqlite_db_path = None, host= None, port= None, database_name= None, 
                  username= None, password=None, collection_name= None, **kwargs):
    """ This function is used to load the data from the database.

    Args:
        database (str): The type of database. For instance 'sqlite', 'mysql', 'postgres', 'mongo'.
        query (str): The query to execute.
        sqlite_db_path (str): The path to the SQLite database.

        The following arguments are used for MySQL, PostgreSQL and MongoDB databases:
        host (str): The host name.
        port (int): The port number.
        database_name (str): The database name.
        username (str): The username if exists. by default None.
        password (str): The password if exists. by default None.
        collection_name (str): The collection name for mongodb database. by default None.
        **kwargs: Keyword arguments to pass to sqlite3.connect, psycopg2.connect, mysql.connector.connect, pymongo.MongoClient.

    Returns:
        pandas.DataFrame: The DataFrame containing the data.
    """
    try:
        if database_type == 'sqlite':
            df = read_sqlite(query, sqlite_db_path)
        elif database_type == 'postgres':
            df = read_postgres(query, host, port, database_name, username, password)
        elif database_type == 'mysql':
            df = read_mysql(query, host, port, database_name, username, password)
        elif database_type == 'mongo':
            df = read_mongo(query, host, port, database_name, username, password, collection_name)
        else:
            raise ValueError('The database type is not supported.')
        
        return df
        
    except ValueError as e:
        print(e)
    except Exception as e:
        print(e)


#Stage 2: Data Preprocessing

def preprocessing(dataframe, label_column,
                missing= {'type': 'drop',
                        'drop_columns': True, 
                        'threshold': 0.25, 
                        'strategy_numerical': 'knn',
                        'strategy_categorical': 'most_frequent',
                        'fill_value': None}, 
                outlier= {'method': 'None',
                        'zscore_threshold': 3,
                        'iqr_threshold': 1.5,
                        'Lc': 0.95, 
                        'Uc': 0.05,
                        'cap': False},
                **kwargs):
    
    """ This function is used to preprocess the data.

    Args:
        dataframe (pandas.DataFrame): The DataFrame containing the data.
        label_column (str): The name of the label column.

        missing (dict): The dictionary containing the imputation parameters. 
        The following parameters are supported:
        
        type (str): The type of imputation. Can be one of "drop", "impute".
        drop_columns (bool): If True, the columns with missing values greater than the threshold will be dropped.
        threshold (float): The threshold to use for dropping columns with missing values. Must be between 0 and 1.
        strategy_numerical (str): The imputation strategy to use for numerical columns. Can be one of "mean", "median", "most_frequent", "knn" or "constant".
        strategy_categorical (str): The imputation strategy to use for categorical columns. Can be one of "most_frequent", or "constant".
        fill_value (str/int/float): The (str) value to use for imputing missing values when strategy is "constant for object data types, For numerical it should be a int/float".

        outlier (dict): The dictionary containing the outlier parameters.
        The following parameters are supported:
        method (str): The method to use for removing outliers. Can be one of "zscore", "iqr", "None".
        iqr_threshold (float): The threshold to use for removing outliers. Must be greater than 0.
        Lc (float): The lower cut-off value to use for removing outliers. Must be between 0 and 1.
        Uc (float): The upper cut-off value to use for removing outliers. Must be between 0 and 1.
        cap (bool): If True, the outliers will be capped to the upper and lower bounds. If False, the outliers will be removed.
        **kwargs: Keyword arguments to pass to sklearn KNNImputer.

    Returns:
        pandas.DataFrame: The DataFrame containing the preprocessed data.
    """

    data = dataframe.copy()

    if missing['drop_columns']:
        data = drop_missing_columns(data, threshold=missing['threshold'])

    if data[label_column].isnull().sum() > 0:
        print('The Label column has missing values. so dropping the rows with missing values in label column')
        data = drop_missing_rows(data,subset=[label_column])


    # Separate columns into numerical and categorical
    numerical_columns = data.select_dtypes(exclude='object').columns
    categorical_columns = data.select_dtypes(include='object').columns

    # Remove label column from numerical and categorical columns
    if data[label_column].dtype == 'object':
        categorical_columns = categorical_columns.drop(label_column)
    else:
        numerical_columns = numerical_columns.drop(label_column)

    if missing['type'] == 'drop':
        data = drop_missing_rows(data)

    elif missing['type'] == 'impute':
        data =  imputation(data, strategy_numerical=missing['strategy_numerical'], 
                            strategy_categorical=missing['strategy_categorical'],
                            fill_value=missing['fill_value'],
                            numerical_columns=numerical_columns,
                            categorical_columns=categorical_columns, 
                            **kwargs)
    if outlier['method'] == 'None':
        pass
    elif outlier['method'] == 'zscore':
        data = remove_outliers_zscore(data, numerical_columns=numerical_columns, zscore_threshold=outlier['zscore_threshold'])
    elif outlier['method'] == 'iqr':
        data = inter_quartile_range(data, numerical_columns=numerical_columns, iqr_threshold=outlier['iqr_threshold'], Lc=outlier['Lc'], Uc=outlier['Uc'], cap=outlier['cap'])
        
    return data

#Stage 3: Data Transformation

def transformation(dataframe, label_column, type = 'ordinal',target_transform = False, cardinality = True, Cardinality_threshold = 0.3):
    """ This function is used to transform the feature data.

    Args:
        dataframe (pandas.DataFrame): The DataFrame containing the data.
        label_column (str): The name of the label column.
        type (str): The type of transformation. Can be one of "ordinal", "onehot".
        target_transform (bool): If True, the label column will be transformed.
        cardinality (bool): If True, the columns with cardinality greater than the threshold will be dropped.
        Cardinality_threshold (float): The threshold to use for dropping columns with cardinality. Must be between 0 and 1.

    Returns:
        pandas.DataFrame: The DataFrame containing the transformed data.
    """
    categorical_columns = dataframe.select_dtypes(include=['object']).columns
    categorical_columns = categorical_columns.drop([label_column]) if label_column in categorical_columns else categorical_columns


    if cardinality:
        drop_columns = [x for x in categorical_columns if dataframe[x].nunique() > Cardinality_threshold*100]
        # drop_columns = [x for x in drop_columns if x not in label_column or x in categorical_columns]
        print('List of columns dropped due to high cardinality: {}'.format(drop_columns))
        dataframe.drop(drop_columns, axis=1, inplace=True)
        categorical_columns = categorical_columns.drop(drop_columns)

    if target_transform:
        dataframe, target_encoder = label_transform(dataframe, label_column)

    
    if type == 'ordinal':
        dataframe, encoder = ordinal_transform(dataframe, categorical_columns)
    elif type == 'onehot':
        dataframe, encoder = onehot_transform(dataframe, categorical_columns)
    elif type == 'None':
        print('No transformation applied on features')
    else:
        raise ValueError('The transformation type is not supported.')
    
    return dataframe, encoder

#Stage 4: Data Scaling

def scaling(dataframe, label_column, type = 'standard', target_scaling = False):
    """ This function is used to scale the feature data.

    Args:
        dataframe (pandas.DataFrame): The DataFrame containing the data.
        label_column (str): The name of the label column.
        type (str): The type of scaling. Can be one of "standard", "robust", "maxabs", "power", "quantile".
        target_scaling (bool): If True, the label column will be scaled using Sklearn MinMax Scaler. Applicable only for Regression usecase.

    Returns:
        pandas.DataFrame: The DataFrame containing the scaled data.
    """
    numerical_columns = dataframe.select_dtypes(exclude='object').columns
    numerical_columns = numerical_columns.drop([label_column]) if label_column in numerical_columns else numerical_columns

    if target_scaling:
        dataframe, target_scaler = target_scale(dataframe, label_column)

    if type == 'standard':
        dataframe, scaler = standard_scale(dataframe, numerical_columns)
    elif type == 'robust':
        dataframe, scaler = robust_scale(dataframe, numerical_columns)
    elif type == 'maxabs':
        dataframe, scaler = maxabs_scale(dataframe, numerical_columns)
    elif type == 'power':
        dataframe, scaler = power_transform(dataframe, numerical_columns)
    elif type == 'quantile':
        dataframe, scaler = quantile_transform(dataframe, numerical_columns)
    elif type == 'None':
        print('No scaling applied on features')
    else:
        raise ValueError('The scaling type is not supported.')
    
    if target_scaling:
        return dataframe, scaler, target_scaler
    else:
        return dataframe, scaler

#Stage 5: Feature Engineering


#Stage 6: Feature Selection
def feature_selection(dataframe, label_column, estimator, type='rfecv', max_features=10, min_features=2, scoring= 'accuracy', 
                    cv=5, n_jobs= -1, **kwargs):
    """ This function is used to select the features.

    Args:
        dataframe (pandas.DataFrame): The DataFrame containing the data.
        label_column (str): The name of the label column.
        estimator (str): The estimator to use for feature selection. Can be one of SVC(),SVR(), RandomForestClassifier(),
        RandomForestRegressor(), GradientBoostingClassifier(), GradientBoostingRegressor(), LogisticRegression(), LinearRegression(), etc..
        type (str): The type of feature selection. Can be one of "rfe", "rfecv".
        max_features (int): The maximum number of features to select.
        min_features (int): The minimum number of features to select.
        scoring (str): The scoring function to use for feature selection. Can be one of "accuracy", "f1", "precision", "recall", "roc_auc", "neg_mean_squared_error", "neg_mean_absolute_error", "neg_median_absolute_error", "r2".
        cv (int): The number of folds to use for cross-validation.
        n_jobs (int): The number of jobs to run in parallel.
        **kwargs: Keyword arguments to pass to sklearn.feature_selection.RFE or sklearn.feature_selection.RFECV.

    Returns:
        pandas.DataFrame: The DataFrame containing the selected features.
    """
    if type == 'rfe':
        df, selector = rfe(dataframe, label_column, estimator, n_features_to_select=max_features, step=1, verbose=0, **kwargs)
    elif type == 'rfecv':
        df, selector = rfecv(dataframe, label_column, estimator, scoring, cv, n_jobs, **kwargs)
    elif type == 'None':
        print('No feature selection applied')
        return None
    else:
        raise ValueError('The feature selection type is not supported.')
    
    return df, selector

#Stage 7: Model Training

def training(dataframe, label_column, model_name, problem_type, target_scaler=None, test_split =0.15, hypertune=True, n_epochs = 100):
    """ This function is used to train the model.

    Args:
        dataframe (pandas.DataFrame): The DataFrame containing the data.
        label_column (str): The name of the label column.
        model_name (str): The name of the model. Can be one of "LinearRegression", "BayesianRidge", "RandomForestRegressor", "XGBRegressor",
                            "GradientBoostingRegressor", "SVR", "DecisionTreeRegressor", "KNeighborsRegressor", "LogisticRegression", 
                            "RandomForestClassifier", "XGBClassifier", "GradientBoostingClassifier", "SVC", "DecisionTreeClassifier", 
                            "KNeighborsClassifier", "GaussianNB".

        problem_type (str): The type of problem. Can be one of "Regression", "Classification".
        target_scaler : The scaler object name used for scaling the label column dduring scaling phase. Applicable only for Regression usecase. 
        test_split (float): The percentage of data to use for testing. it should be 0 to 1.
        hypertune (bool): If True, the model will be hypertuned using Optuna.
        n_epochs (int): The number of epochs to use for hypertuning.

    Returns:
        sklearn.pipeline.Pipeline: The pipeline containing the trained model.
    """
    # Split the dataset into train and test
    x_train, x_test, y_train, y_test = train_test_split(dataframe.drop(label_column, axis=1), dataframe[label_column], 
                                                        test_size=test_split, random_state=42)

    # with mlflow.start_run(run_name=datetime.now().strftime('%Y-%m-%d_%H:%M:%S')) as run:
    #     run_id = run.info.run_id
        # print('MLflow Run ID: {}'.format(run_id))
        # client = MlflowClient()
        # client.set_tag(run_id, "mlflow.note.content", f"ML model: {model_name}")

        # tags = {"Application": "AutoPilotML"}
        # mlflow.set_tags(tags)

    # Log the model
    if "Xgboost" in model_name:
        mlflow.xgboost.autolog(log_input_examples=True, log_model_signatures=True, log_models=True, 
        disable=False, exclusive=False, disable_for_unsupported_versions=False, silent=False,log_post_training_metrics=True
        )
    else:
        mlflow.sklearn.autolog(log_input_examples=True, log_model_signatures=True, log_models=True, 
        disable=False, exclusive=False, disable_for_unsupported_versions=False, silent=False,log_post_training_metrics=True
        )


    if hypertune:

        def objective(trial):
            # best_accuracy = float('inf') if problem_type == 'regression' else 0.0
            # best_model = None

            tuner_params = tuner(model_name, trial)

            model, y_pred = model_fitting(x_train, x_test, y_train, y_test, model_name= model_name, params= tuner_params)
            
            if problem_type == 'Regression':
                if target_scaler is not None:
                    y_pred = target_scaler.inverse_transform(y_pred.reshape(-1,1))
                accuracy = round(mean_squared_error(y_test, y_pred, squared=False),4)
            else:
                accuracy = round(accuracy_score(y_test, y_pred),4)

            return accuracy

        study = optuna.create_study(direction='minimize' if problem_type == 'Regression' else 'maximize')
        study.optimize(objective, n_trials=n_epochs)

        best_params = study.best_params

        print('Best Score: {}'.format(study.best_value))
        print('Best Parameters: {}'.format(study.best_params))

        model, y_pred = model_fitting(x_train, x_test, y_train, y_test, model_name= model_name, params= best_params)

    else:
        model, y_pred = model_fitting(x_train, x_test, y_train, y_test, model_name= model_name)

        if problem_type == 'Regression':
            if target_scaler is not None:
                y_pred = target_scaler.inverse_transform(y_pred.reshape(-1,1))
            accuracy = round(mean_squared_error(y_test, y_pred, squared=False),4)
        else:
            accuracy = round(accuracy_score(y_test, y_pred),4)
        
        print('Accuracy: {}'.format(accuracy))

    return model



























import numpy as np

def inter_quartile_range(df, iqr_threshold, Lc, Uc, cap, numerical_columns, **kwargs):

    """ This function is used to remove outliers from the dataset using the inter-quartile range method.
    Args:
        df (pandas.DataFrame): The DataFrame containing the outliers.
        iqr_threshold (float): The threshold to use for removing outliers. Must be greater than 0.
        Lc (float): The lower cut-off value to use for removing outliers. Must be between 0 and 1.
        Uc (float): The upper cut-off value to use for removing outliers. Must be between 0 and 1.
        cap (bool): If True, the outliers will be capped to the upper and lower bounds. If False, the outliers will be removed.
    Returns:
        pandas.DataFrame: The DataFrame containing the outliers.
    """

    for col in numerical_columns:
        Q1 = df[col].quantile(Lc)
        Q3 = df[col].quantile(Uc)

        IQR = Q3 - Q1
        lower_bound = Q1 - iqr_threshold*IQR
        upper_bound = Q3 + iqr_threshold*IQR

        if cap:
            df[col] = df[col].apply(lambda x: upper_bound if x > upper_bound else lower_bound if x < lower_bound else x)
        else:
            df = df[(df[col] >= lower_bound) & (df[col] <= upper_bound)]

    return df


def remove_outliers_zscore(df, numerical_columns, zscore_threshold=3, **kwargs):

    """ This function is used to remove outliers from the dataset using the z-score method.
    Args:
        df (pandas.DataFrame): The DataFrame containing the outliers.
        remove_outliers_zscore_threshold (float): The threshold to use for removing outliers. Must be greater than 0.
    Returns:
        pandas.DataFrame: The DataFrame containing the outliers.
    """

    outlier_mask = np.abs((df[numerical_columns] - df[numerical_columns].mean()) / df[numerical_columns].std()) > zscore_threshold
    df = df[~outlier_mask.any(axis=1)]

    return df
























































































































from sklearn.linear_model import LinearRegression, BayesianRidge, LogisticRegression
from sklearn.ensemble import RandomForestRegressor, RandomForestClassifier, GradientBoostingRegressor, GradientBoostingClassifier
from sklearn.tree import DecisionTreeRegressor, DecisionTreeClassifier
from sklearn.neighbors import KNeighborsRegressor, KNeighborsClassifier
from sklearn.svm import SVR, SVC
from sklearn.naive_bayes import GaussianNB
from xgboost import XGBRegressor, XGBClassifier


# Regression Models

def model_fitting(x_train, x_test, y_train, y_test, model_name, params):
    """
    This function is used to create a Regression model.

    Args:
        x_train (pandas.DataFrame): The DataFrame containing the training data.
        x_test (pandas.DataFrame): The DataFrame containing the testing data.
        y_train (pandas.Series): The Series containing the training labels.
        y_test (pandas.Series): The Series containing the testing labels.
        model_name (str): The name of the model.
        **kwargs: The hyperparameters for the selected Regression model.

    Returns:
        model : The selected Regression model.
        y_pred (numpy.ndarray): The predicted labels.
    """
    model_list = {
        # Regression models
        'LinearRegression': LinearRegression,
        'BayesianRidge': BayesianRidge,
        'RandomForestRegressor': RandomForestRegressor,
        'XGBRegressor': XGBRegressor,
        'GradientBoostingRegressor': GradientBoostingRegressor,
        'SVR': SVR,
        'DecisionTreeRegressor': DecisionTreeRegressor,
        'KNeighborsRegressor': KNeighborsRegressor,

        # Classifier models
        'LogisticRegression': LogisticRegression,
        'RandomForestClassifier': RandomForestClassifier,
        'XGBClassifier': XGBClassifier,
        'GradientBoostingClassifier': GradientBoostingClassifier,
        'SVC': SVC,
        'DecisionTreeClassifier': DecisionTreeClassifier,
        'KNeighborsClassifier': KNeighborsClassifier,
        'GaussianNB': GaussianNB
    }

    if model_name not in model_list:
        raise ValueError("Model not supported")

    # Create the selected model
    model_class = model_list[model_name]
    
    model = model_class(**params)

    # Fit the model
    model.fit(x_train, y_train)

    # Predict the labels
    y_pred = model.predict(x_test)

    return model, y_pred





import numpy as np
import pandas as pd
from sklearn.preprocessing import LabelEncoder, OneHotEncoder, OrdinalEncoder

def onehot_transform(df, categorical_columns):
    """
    This function is used to transform categorical data into numerical data.

    Args:
        df (pandas.DataFrame): The DataFrame containing the data.
        categorical_columns (list): The list of categorical columns.
        
    Returns:
        pandas.DataFrame: The DataFrame containing the transformed data.
        encoder (sklearn.preprocessing.OneHotEncoder): The OneHotEncoder object used for transformation.
    """

    # OneHotEncoder for binary categorical data
    encoder = OneHotEncoder(handle_unknown ='ignore')
    feature_array = encoder.fit_transform(df[categorical_columns]).toarray()
    feature_labels = encoder.get_feature_names_out(categorical_columns)
    df[feature_labels] = feature_array
    df.drop(categorical_columns, axis=1, inplace=True)

    return df, encoder

def ordinal_transform(df, categorical_columns):
    """
    This function is used to transform categorical data into numerical data.

    Args:
        df (pandas.DataFrame): The DataFrame containing the data.
        categorical_columns (list): The list of categorical columns.

    Returns:
        pandas.DataFrame: The DataFrame containing the transformed data.
        encoder (sklearn.preprocessing.OrdinalEncoder): The OrdinalEncoder object used for transformation.
    """

    # OrdinalEncoder for binary categorical data
    encoder = OrdinalEncoder()
    feature_array = encoder.fit_transform(df[categorical_columns])
    df[categorical_columns] = feature_array

    return df, encoder

def label_transform(df, target_column):
    """
    This function is used to transform target feature into numerical feature.

    Args:
        df (pandas.DataFrame): The DataFrame containing the data.
        target_column (str): The name of the target column.

    Returns:
        pandas.DataFrame: The DataFrame containing the transformed data.
        encoder (sklearn.preprocessing.LabelEncoder): The LabelEncoder object used for transformation.
    """

    # OrdinalEncoder for binary categorical data
    encoder = LabelEncoder()
    df[target_column] = encoder.fit_transform(df[target_column])

    return df, encoder


import optuna

def tuner(model_name, trial: optuna.Trial):
    """
    This function is used to tune the hyperparameters of the selected model.

    Args:
        model_name (str): The name of the model.
        trial (optuna.trial): The trial object that stores the hyperparameters.

    Returns:
        tuner_params (dict): The hyperparameters for the selected model.
    """

    tuner_params = {
        'LinearRegression': lambda:{
            'fit_intercept': trial.suggest_categorical("fit_intercept", ["True", "False"]), 
            'positive': trial.suggest_categorical("positive", ["True", "False"])
                            },

        'BayesianRidge': lambda:{
            'max_iter': trial.suggest_int("max_iter", 50, 1000), 
            'tol': trial.suggest_float("tol", 1e-9, 1.0, log=True),
            'lambda_1': trial.suggest_float("lambda_1", 1e-9, 1.0, log=True),
            'lambda_2': trial.suggest_float("lambda_2", 1e-9, 1.0, log=True),
            'alpha_1': trial.suggest_float('alpha_1', 1e-8, 1.0, log=True), 
            'alpha_2': trial.suggest_float('alpha_2', 1e-8, 1.0, log=True),
            'compute_score': trial.suggest_categorical("compute_score", ["True", "False"]),
            'fit_intercept': trial.suggest_categorical("fit_intercept", ["True", "False"])
                        },

        'RandomForestRegressor': lambda: {
            'n_estimators': trial.suggest_int('n_estimators', 100, 1000, step=100),
            'max_depth': trial.suggest_int('max_depth', 3, 10, step=1),
            'min_samples_split': trial.suggest_int('min_samples_split', 2, 10, step=1),
            'min_samples_leaf': trial.suggest_int('min_samples_leaf', 1, 10, step=1),
            'max_features': trial.suggest_categorical('max_features', ['auto', 'sqrt', 'log2']),
            'bootstrap': trial.suggest_categorical('bootstrap', ['True', 'False'])
                                },

        'XGBRegressor': lambda: {
            'n_estimators': trial.suggest_int('n_estimators', 100, 1000, step=100),
            'max_depth': trial.suggest_int('max_depth', 3, 10, step=1),
            'learning_rate': trial.suggest_float('learning_rate', 0.001, 0.1, log=True),
            'subsample': trial.suggest_float('subsample', 0.6, 1.0, log=True),
            'colsample_bytree': trial.suggest_float('colsample_bytree', 0.6, 1.0, log=True),
            'reg_alpha': trial.suggest_float('reg_alpha', 1e-8, 100.0, log=True),
            'reg_lambda': trial.suggest_float('reg_lambda', 1e-8, 100.0, log=True),
            'min_child_weight': trial.suggest_int('min_child_weight', 1, 10, step=1),
            'gamma': trial.suggest_float('gamma', 1e-8, 1.0, log=True),
        },

        'GradientBoostingRegressor': lambda: {
            'n_estimators': trial.suggest_int('n_estimators', 50, 500, step=50),
            'learning_rate': trial.suggest_float('learning_rate', 0.01, 0.1),
            'max_depth': trial.suggest_int('max_depth', 3, 10, step=1),
            'min_samples_split': trial.suggest_int('min_samples_split', 2, 20, step=1),
            'min_samples_leaf': trial.suggest_int('min_samples_leaf', 1, 20, step=1),
            'max_features': trial.suggest_categorical('max_features', ['auto', 'sqrt', 'log2'])
            },

        'SVR': lambda: {
            'kernel': trial.suggest_categorical('kernel', ['linear', 'rbf', 'poly', 'sigmoid']),
            'C': trial.suggest_float('C', 1e-3, 1e3, log=True),
            'gamma': trial.suggest_float('gamma', 1e-3, 1e3, log=True) if ['kernel'] in ['rbf', 'poly'] else 'scale',
            'degree': trial.suggest_int('degree', 2, 5, step=1) if ['kernel'] == 'poly' else 3
            },

        'DecisionTreeRegressor': lambda: {
            'max_depth': trial.suggest_int('max_depth', 2, 32, step=1),
            'min_samples_split': trial.suggest_int('min_samples_split', 2, 20, step=1),
            'min_samples_leaf': trial.suggest_int('min_samples_leaf', 1, 20, step=1),
            'max_features': trial.suggest_categorical('max_features', ['auto', 'sqrt', 'log2'])
            },

        'KNeighborsRegressor': lambda: {
            'n_neighbors': trial.suggest_int('n_neighbors', 1, 20, step=1),
            'weights': trial.suggest_categorical('weights', ['uniform', 'distance']),
            'p': trial.suggest_int('p', 1, 2),  # Minkowski distance power parameter
            'leaf_size': trial.suggest_int('leaf_size', 1, 50, step=1),
            'algorithm': trial.suggest_categorical('algorithm', ['auto', 'ball_tree', 'kd_tree', 'brute'])

            },
        'LogisticRegression': lambda: {
            'C': trial.suggest_float('C', 1e-5, 1e2, log=True),
            'penalty': trial.suggest_categorical('penalty', ['l1', 'l2', 'None']),
            'solver': trial.suggest_categorical('solver', ['liblinear', 'lbfgs', 'newton-cg', 'sag', 'saga']),
            'max_iter': trial.suggest_int('max_iter', 100, 1000, step=100),
            },

        'RandomForestClassifier': lambda: {
            'n_estimators': trial.suggest_int('n_estimators', 100, 1000, step=100),
            'max_depth': trial.suggest_int('max_depth', 3, 10, step=1),
            'min_samples_split': trial.suggest_int('min_samples_split', 2, 20, step=1),
            'min_samples_leaf': trial.suggest_int('min_samples_leaf', 1, 10, step=1),
            'max_features': trial.suggest_categorical('max_features', ['auto', 'sqrt', 'log2']),
            'criterion': trial.suggest_categorical('criterion', ['gini', 'entropy']),
            'bootstrap': trial.suggest_categorical('bootstrap', ['True', 'False'])
            },

        'XGBClassifier':  lambda: {
            'max_depth': trial.suggest_int('max_depth', 3, 10, step=1),
            'learning_rate': trial.suggest_float('learning_rate', 0.01, 0.1, step=0.01),
            'n_estimators': trial.suggest_int('n_estimators', 100, 1000, step=100),
            'subsample': trial.suggest_float('subsample', 0.6, 1.0, step=0.1),
            'colsample_bytree': trial.suggest_float('colsample_bytree', 0.6, 1.0, step=0.1),
            'gamma': trial.suggest_float('gamma', 0.0, 1.0, step=0.1),
            'min_child_weight': trial.suggest_int('min_child_weight', 1, 10, step=1),
            'reg_alpha': trial.suggest_float('reg_alpha', 0.0, 1.0, step=0.1),
            'reg_lambda': trial.suggest_float('reg_lambda', 0.0, 1.0, step=0.1),
            },

        'GradientBoostingClassifier':  lambda: {
            'learning_rate': trial.suggest_float('learning_rate', 0.001, 0.5, step=0.05),
            'n_estimators': trial.suggest_int('n_estimators', 100, 1000, step=100),
            'max_depth': trial.suggest_int('max_depth', 3, 10, step=1),
            'min_samples_split': trial.suggest_int('min_samples_split', 2, 10, step=1),
            'min_samples_leaf': trial.suggest_int('min_samples_leaf', 1, 5, step=1),
            'subsample': trial.suggest_float('subsample', 0.5, 1.0, step=0.1),
            },

        'SVC': lambda: {
            'C': trial.suggest_float('C', 1e-5, 1e2, log=True),  # Regularization parameter
            'gamma': trial.suggest_float('gamma', 1e-5, 1e2, log=True),  # Kernel coefficient
            'kernel': trial.suggest_categorical('kernel', ['linear', 'rbf', 'poly']),  # Kernel type
            'degree': trial.suggest_int('degree', 2, 5, step=1) if ['kernel'] == 'poly' else 3  # Degree of the polynomial kernel function
            },

        'DecisionTreeClassifier': lambda: {
            'max_depth': trial.suggest_int('max_depth', 2, 32, step=2),
            'min_samples_split': trial.suggest_int('min_samples_split', 2, 20, step=2),
            'min_samples_leaf': trial.suggest_int('min_samples_leaf', 1, 20, step=1),
            'max_features': trial.suggest_categorical('max_features', ['auto', 'sqrt', 'log2', None]),
            'criterion': trial.suggest_categorical('criterion', ['gini', 'entropy'])
            },

        'KNeighborsClassifier': lambda: {
            'n_neighbors': trial.suggest_int('n_neighbors', 1, 10, step=1),
            'weights': trial.suggest_categorical('weights', ['uniform', 'distance']),
            'p': trial.suggest_int('p', 1, 2),  # For Minkowski distance (p=1 for Manhattan, p=2 for Euclidean)
            },

        'GaussianNB': lambda: {
            'var_smoothing': trial.suggest_float('var_smoothing', 1e-12, 1e-3, log=True)
        }
    }

    if model_name not in tuner_params:
        raise ValueError(f"Model {model_name} is not supported for hyperparameter tuning.")


    # param = tuner_params[model_name]
    
    return tuner_params.get(model_name, lambda: {})()







import numpy as np
import pandas as pd
from sklearn.impute import SimpleImputer, KNNImputer





def drop_missing_rows(df,**kwargs):
        
        """ This function is used to drop rows with missing values greater than the threshold.
        Args:
            df (pandas.DataFrame): The DataFrame containing the missing values.
            threshold (float): The threshold to use for dropping rows with missing values. Must be between 0 and 1.
        Returns:
            pandas.DataFrame: The DataFrame containing the dropped rows.
        """

        # Drop rows with missing values greater than the threshold
        df = df.dropna(**kwargs, axis=0)

        return df

def drop_missing_columns(df, threshold=0.25):

    """ This function is used to drop columns with missing values greater than the threshold.
    Args:
        df (pandas.DataFrame): The DataFrame containing the missing values.
        threshold (float): The threshold to use for dropping columns with missing values. Must be between 0 and 1.
    Returns:
        pandas.DataFrame: The DataFrame containing the dropped columns.
    """

    # Drop columns with missing values greater than the threshold
    df = df.dropna(thresh=threshold*len(df), axis=1)

    return df  

def imputation(df, strategy_numerical, strategy_categorical, numerical_columns, categorical_columns, fill_value = None, **kwargs):

    """ This function is used to impute the missing values in the dataset.
    Args:
        df (pandas.DataFrame): The DataFrame containing the missing values.
        strategy_numerical (str): The imputation strategy to use for numerical columns. Can be one of "mean", "median", "most_frequent", "knn" or "constant".
        strategy_categorical (str): The imputation strategy to use for categorical columns. Can be one of "most_frequent", or "constant".
        fill_value (str/int/float): The (str) value to use for imputing missing values when strategy is "constant for object data types, For numerical it should be a int/float".
    Returns:
        pandas.DataFrame: The DataFrame containing the imputed data.
    """

    if numerical_columns.any():
        
        if strategy_numerical == 'knn':
            # Impute missing values for numerical columns using KNN
            # imputer = KNNImputer(n_neighbors=kwargs.get('n_neighbors', 5), weights=kwargs.get('weights', 'uniform'))
            imputer = KNNImputer(**kwargs)
            df[numerical_columns] = imputer.fit_transform(df[numerical_columns])

        else:
            # Impute missing values for numerical columns using SimpleImputer
            imputer = SimpleImputer(strategy=strategy_numerical, fill_value= [None if strategy_numerical != 'constant' else fill_value])
            df[numerical_columns] = imputer.fit_transform(df[numerical_columns])

    if categorical_columns.any():
        # Impute missing values for categorical columns using SimpleImputer
        imputer = SimpleImputer(strategy=strategy_categorical, fill_value=[None if strategy_numerical != 'constant' else fill_value])
        df[categorical_columns] = imputer.fit_transform(df[categorical_columns])

    return df







import pandas as pd
import numpy as np
from sqlite3 import connect



def read_csv(path, **kwargs):
    """Reads a CSV file and returns a pandas DataFrame.

    Args:
        path (str): Path to the CSV file.
        **kwargs: Keyword arguments to pass to pandas.read_csv.

    Returns:
        pandas.DataFrame: The DataFrame containing the data from the CSV file.
    """
    return pd.read_csv(path, **kwargs)

def read_excel(path, **kwargs):
    """Reads an Excel file and returns a pandas DataFrame.

    Args:
        path (str): Path to the Excel file.
        **kwargs: Keyword arguments to pass to pandas.read_excel.

    Returns:
        pandas.DataFrame: The DataFrame containing the data from the Excel file.
    """
    return pd.read_excel(path, **kwargs)




import pathlib
import setuptools

setuptools.setup(
    name="autopilotml",
    version="1.0.1",
    keywords="autopilot, autopilotml, autopilot-ml, autopilot ml",
    author="Shyam Prasath",
    author_email="shshyam96@gmail.com",
    description="A package for automating machine learning tasks",
    long_description=pathlib.Path("README.md").read_text(),
    long_description_content_type="text/markdown",
    url="https://github.com/shyam1326",
    packages=setuptools.find_packages(),
    include_package_data=True,
    license="MIT License (MIT)",
    project_urls={
        "Source": "https://github.com/shyam1326/autopilotml",
        "Bug Reports": "https://github.com/shyam1326/autopilotml/issues",
        "Documentation": "https://github.com/shyam1326/autopilotml/blob/main/README.md",
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: Apache Software License",
    ],
    python_requires=">=3.8, <3.12",
    install_requires=[requirement.strip() for requirement in pathlib.Path("requirements.txt").read_text().splitlines()],
    entry_points={
        "console_scripts": [
            "autopilotml=autopilotml.__main__:autopilotml",
        ],
    },
    

)



from sklearn.preprocessing import StandardScaler, MinMaxScaler, RobustScaler, MaxAbsScaler, PowerTransformer, QuantileTransformer


def standard_scale(df, feature_columns):
    """
    This function is used to scale the numerical data using StandardScaler.

    Args:
        df (pandas.DataFrame): The DataFrame containing the data.
        feature_columns (list): The list of feature columns exclude the target column.

    Returns:
        pandas.DataFrame: The DataFrame containing the scaled data.
        scaler (sklearn.preprocessing.StandardScaler): The StandardScaler object used for scaling.
    """

    # StandardScaler for numerical data
    scaler = StandardScaler()
    df[feature_columns] = scaler.fit_transform(df[feature_columns])

    return df, scaler

def target_scale(df, target_column):
    """
    This function is used to scale the numerical data using StandardScaler.

    Args:
        df (pandas.DataFrame): The DataFrame containing the data.
        target_column (str): The name of the target column.

    Returns:
        pandas.DataFrame: The DataFrame containing the scaled data.
        scaler (sklearn.preprocessing.MinMaxScaler): The MinMaxScaler object used for scaling.
    """

    # StandardScaler for numerical data
    scaler = MinMaxScaler()
    df[target_column] = scaler.fit_transform(df[target_column])

    return df, scaler

def robust_scale(df, feature_columns):
    """
    This function is used to scale the numerical data using RobustScaler.

    Args:
        df (pandas.DataFrame): The DataFrame containing the data.
        feature_columns (list): The list of feature columns exclude the target column.


    Returns:
        pandas.DataFrame: The DataFrame containing the scaled data.
        scaler (sklearn.preprocessing.RobustScaler): The RobustScaler object used for scaling.
    """

    # RobustScaler for numerical data
    scaler = RobustScaler()
    df[feature_columns] = scaler.fit_transform(df[feature_columns])

    return df, scaler

def maxabs_scale(df, feature_columns):
    """
    This function is used to scale the numerical data using MaxAbsScaler.

    Args:
        df (pandas.DataFrame): The DataFrame containing the data.
        feature_columns (list): The list of feature columns exclude the target column.

    Returns:
        pandas.DataFrame: The DataFrame containing the scaled data.
        scaler (sklearn.preprocessing.MaxAbsScaler): The MaxAbsScaler object used for scaling.
    """

    # MaxAbsScaler for numerical data
    scaler = MaxAbsScaler()
    df[feature_columns] = scaler.fit_transform(df[feature_columns])

    return df, scaler

def power_transform(df, feature_columns, method='yeo-johnson'):
    """
    This function is used to scale the numerical data using PowerTransformer.

    Args:
        df (pandas.DataFrame): The DataFrame containing the data.
        feature_columns (list): The list of feature columns exclude the target column.
        method (str): The method to use for transformation. Can be one of "yeo-johnson", "box-cox".

    Returns:
        pandas.DataFrame: The DataFrame containing the scaled data.
        scaler (sklearn.preprocessing.PowerTransformer): The PowerTransformer object used for scaling.
    """

    # PowerTransformer for numerical data
    scaler = PowerTransformer(method=method)
    df[feature_columns] = scaler.fit_transform(df[feature_columns])

    return df, scaler

def quantile_transform(df, feature_columns, method='uniform'):
    """
    This function is used to scale the numerical data using QuantileTransformer.

    Args:
        df (pandas.DataFrame): The DataFrame containing the data.
        feature_columns (list): The list of feature columns exclude the target column.
        method (str): The method to use for transformation. Can be one of "uniform", "normal".

    Returns:
        pandas.DataFrame: The DataFrame containing the scaled data.
        scaler (sklearn.preprocessing.QuantileTransformer): The QuantileTransformer object used for scaling.
    """

    # QuantileTransformer for numerical data
    scaler = QuantileTransformer(method=method)
    df[feature_columns] = scaler.fit_transform(df[feature_columns])

    return df, scaler